// IEEEFloat.cpp: Inspects bits of a float for a little-endian architecture
#include <bitset>
#include <climits>
#include <cmath>
#include <iostream>
#include <limits>
#include <vector>
using namespace std;

void ieeePrint(float x) {
    unsigned char* p = reinterpret_cast<unsigned char*>(&x);
    int bitno = 0;
        
    for (int i = sizeof(float)-1; i >= 0; --i) {
        for (int j = CHAR_BIT-1; j >= 0; --j, ++bitno) {
            cout << !!(p[i] & (1 << j));
            if (bitno == 0 || bitno == 8)  // IEEE boundaries
                cout << ' ';
        }
    }
    cout << endl;
}

void printBitsOfByte(unsigned char byte) {
   union bits {
      unsigned char byte;
      struct {
         unsigned bit7 : 1;
         unsigned bit6 : 1;
         unsigned bit5 : 1;
         unsigned bit4 : 1;
         unsigned bit3 : 1;
         unsigned bit2 : 1;
         unsigned bit1 : 1;
         unsigned bit0 : 1;
      };
   };
   bits b;
   b.byte = byte;
   cout << b.bit7 << b.bit6 << b.bit5 << b.bit4
        << b.bit3 << b.bit2 << b.bit1 << b.bit0;
}

void printBytes(float x) {
   union convert {
      unsigned char bytes[sizeof(float)];
      float f;
   } c;
   c.f = x;
   for (int i = 0; i < sizeof(float); ++i) 
      printBitsOfByte(c.bytes[i]);
   cout << endl;
}

int main() {
    cout << boolalpha<< numeric_limits<float>::is_iec559
         << noboolalpha << endl;
    ieeePrint(1.0);
    ieeePrint(2.0);
    ieeePrint(6.5);
    ieeePrint(-6.5);
    ieeePrint(-27.5f);
    ieeePrint(-361.0f);
    ieeePrint(-236.0f);
    ieeePrint(12.5f);
    ieeePrint(0.0);
    float inf = numeric_limits<float>::infinity();
    ieeePrint(inf);
    float zero = 1.0f/inf;
    ieeePrint(zero);
    cout << "1/inf = " << zero << endl;
    float negzero = 1/-inf;
    ieeePrint(negzero);
    cout << "1/-inf = " << negzero << endl;
    float x = -4.0f;
    float nan = sqrt(x);
    ieeePrint(nan);
    ieeePrint(nan+2);
    ieeePrint(numeric_limits<float>::quiet_NaN());
    ieeePrint(numeric_limits<float>::signaling_NaN());
    cout << boolalpha << (nan == nan) << endl << noboolalpha;
    ieeePrint(numeric_limits<float>::min());
    ieeePrint(numeric_limits<float>::denorm_min());

    // A quick way to get the bits of a float using bitset
    // (Independent of endian-ness)
    x = -6.5f;
    unsigned long* p = reinterpret_cast<unsigned long*>(&x);
    bitset<sizeof(unsigned long)*CHAR_BIT> bs(*p);
    cout << endl << bs << endl;

    // Now reveal endian-ness
    printBytes(x);

    cout << numeric_limits<float>::round_style << endl;
    cout << numeric_limits<float>::traps << endl;
}

/* Output (VC++):
0 01111111 00000000000000000000000
0 10000000 00000000000000000000000
0 10000001 10100000000000000000000
1 10000001 10100000000000000000000
1 10000011 10111000000000000000000
1 10000111 01101001000000000000000
1 10000110 11011000000000000000000
0 10000010 10010000000000000000000
0 00000000 00000000000000000000000
0 11111111 00000000000000000000000
0 00000000 00000000000000000000000
1/inf = 0
1 00000000 00000000000000000000000
1/-inf = 0
1 11111111 10000000000000000000000
1 11111111 10000000000000000000000
0 11111111 10000000000000000000000
0 11111111 10000000000000000000001
false
0 00000001 00000000000000000000000
0 00000000 00000000000000000000001

11000000110100000000000000000000
00000000000000000000101100000011
1
1
*/
