// parms.cpp: Calculate machine parameters the hard
// (but sure and portable) way. You will notice that
// certain temporary variables are used that may not
// appear needed. This is to prevent aggressive
// optimizers from obscuring the results. See ep1
// and epsp1.

#include <iostream>
#include <limits>
using namespace std;

template<typename FType>
void parms() {
   // Find the base, B. To do this, we use the fact
   // that the smallest power, e, for which B^e + 1.0
   // is not representable is where numbers have a
   // spacing of B (the first interval spacing greater
   // than 1; which happens e = p). This is done
   // by setting y to 1.0 and doubling y until we find
   // the greater-than-one gap. (2 is the smallest
   // possible base, so this will do the job). We
   // then find the next floating-point number after
   // y, and subtract to get B.
   FType y = 1, yp1;
   do {
      y *= 2;
      yp1 = y + 1;
   } while (yp1 - y == 1);
   // B^p <= y < B^(p+1)

   // Now, find the next floating-point number past y.
   // To do this, add successive numbers starting with
   // 1 to y. The first time we get a number other
   // than y, it will be the next floating-point
   // number. Subtract y from this to get B.
   FType x = 1;
   do
      x += 1;
   while (y + x == y);
   // y + x rounds to the first fp number past y now.
   int base = int((y + x) - y);
   cout << "base = " << base << endl;

   // Now that we know the base, we can start over
   // and find the precision, p.
   y = 1;
   int p = 0;
   FType ep1;
   do {
      y *= FType(base);
      ++p;
      ep1 = y + 1;
   } while (ep1 - y == 1);

   cout << "p = " << p << endl;

   // At this point we could compute eps, but let's show
   // how to do it cold. We'll add successively smaller
   // negative powers of 2 to 1 until we get 1 right
   // back. The guess before the last (which will be eps)
   // will be >= the true eps. We'll just add this to 1
   // to land on the next floating-point number past 1, and 
   // then subtract 1 to get the difference.
   FType eps, epsp1;
   x = 1;
   do {
      eps = x;
      x = x / 2;
      epsp1 = x + 1;
   } while (epsp1 > 1);
   cout << "eps = " << eps << endl;
   
   // Compare to expected answers
   cout << "-------------------\n";
   cout << "base should be: " << numeric_limits<FType>::radix << endl;
   cout << "p should be: " << numeric_limits<FType>::digits << endl;
   cout << "eps should be: " << numeric_limits<FType>::epsilon() << endl;
}

int main() {
   parms<float>();
   cout << endl;
   parms<double>();
}

/* Output:
base = 2
p = 24
eps = 1.19209e-07
-------------------
base should be: 2
p should be: 24
eps should be: 1.19209e-07

base = 2
p = 53
eps = 2.22045e-16
-------------------
base should be: 2
p should be: 53
eps should be: 2.22045e-16
*/
